mod api;
pub(crate) mod settings;

use std::{
        net::SocketAddr,
        sync::Arc,
        fmt::Display,
};

use tokio::sync::OnceCell;

use axum::{
    routing::{get},
    Router,
    http::StatusCode,
    response::IntoResponse,
};

use crate::settings::PowerDnsOidcTsigkeyConfig;
use reqwest::Response as ReqwestResponse;
use reqwest::Client;
use serde::{de::DeserializeOwned, Deserialize, Serialize};
//use serde_with::{serde_as, DisplayFromStr};

#[derive(Debug)]
pub enum PowerDnsOidcTsigkeyError {
    ApiResponse(String),
    PowerDnsApi(StatusCode, PowerDnsApiError),
    Reqwest(reqwest::Error),
    Config(config::ConfigError),
    Serde(serde_json::Error),
    Io(std::io::Error),
}

impl PowerDnsOidcTsigkeyError {
    pub async fn from_response(res: ReqwestResponse) -> Self {
        match res
            .headers()
            .get(reqwest::header::CONTENT_TYPE)
            .map(reqwest::header::HeaderValue::to_str)
        {
            Some(Ok(content_type)) if content_type.starts_with("application/json") => {
                let status_code = res.status();
                match res.json().await.map_err(reqwest::Error::from) {
                    Ok(json_result) => Self::PowerDnsApi(status_code, json_result),
                    Err(err) => Self::Reqwest(err),
                }
            }
            _ => match res.text().await {
                Ok(text) => Self::ApiResponse(text),
                Err(err) => Self::Reqwest(err),
            }
        }
    }
}

impl IntoResponse for PowerDnsOidcTsigkeyError {
    fn into_response(self) -> axum::response::Response {
        let sc = match self {
            Self::PowerDnsApi(sc, _) => sc,
            _ => StatusCode::INTERNAL_SERVER_ERROR,
        };
        (sc, self.to_string()).into_response()

    }
}

impl Display for PowerDnsOidcTsigkeyError {
    fn fmt(&self, f: &mut std::fmt::Formatter<'_>) -> std::fmt::Result {
        match self {
            PowerDnsOidcTsigkeyError::ApiResponse(err) => f.write_fmt(format_args!("API Error: {:#?}", err)),
            PowerDnsOidcTsigkeyError::PowerDnsApi(sc, err) => f.write_fmt(format_args!("API Error {}: {:#?}", sc, err)),
            PowerDnsOidcTsigkeyError::Reqwest(err) => f.write_fmt(format_args!("Reqwest Error: {:#?}", err)),
            PowerDnsOidcTsigkeyError::Config(err) => f.write_fmt(format_args!("Config Error: {:#?}", err)),
            PowerDnsOidcTsigkeyError::Serde(err) => f.write_fmt(format_args!("JSON Error: {:#?}", err)),
            PowerDnsOidcTsigkeyError::Io(err) => f.write_fmt(format_args!("IO Error: {:#?}", err)),
        }
    }
}

impl std::error::Error for PowerDnsOidcTsigkeyError {}

impl From<reqwest::Error> for PowerDnsOidcTsigkeyError {
    fn from(e: reqwest::Error) -> Self {
        Self::Reqwest(e)
    }
}

impl From<config::ConfigError> for PowerDnsOidcTsigkeyError {
    fn from(e: config::ConfigError) -> Self {
        Self::Config(e)
    }
}

impl From<serde_json::Error> for PowerDnsOidcTsigkeyError {
    fn from(e: serde_json::Error) -> Self {
        Self::Serde(e)
    }
}

impl From<std::io::Error> for PowerDnsOidcTsigkeyError {
    fn from(e: std::io::Error) -> Self {
        Self::Io(e)
    }
}


#[derive(Clone, Debug, Deserialize, Serialize)]
pub struct PowerDnsApiError {
    pub error: String,
    pub errors: Option<Vec<String>>,
}

pub(crate) type PowerDnsOidcTsigkeyResult<T> = Result<T, PowerDnsOidcTsigkeyError>;

async fn parse_json<Out: DeserializeOwned>(res: ReqwestResponse) -> PowerDnsOidcTsigkeyResult<Out> {
    match res.status().is_success() {
        true => Ok(res.json().await?),
        false => Err(PowerDnsOidcTsigkeyError::from_response(res).await),
    }
}


#[derive(Clone, Debug)]
pub struct AppState {
    http_client: Client,
}

static config_cell: OnceCell<PowerDnsOidcTsigkeyConfig> = OnceCell::const_new();

#[tokio::main]
async fn main() {
    match settings::PowerDnsOidcTsigkeyConfig::load("config.yaml") {
        Ok(config) => {
            config_cell.set(config).unwrap();
            run().await;
        },
        Err(e) => println!("Failed to load config.yaml: {:?}", e),
    };
}

async fn run() {
    let addr: SocketAddr = (config_cell.get().unwrap().server.bind_address, config_cell.get().unwrap().server.port).into();
    let state = AppState { http_client: reqwest::Client::new() };
//    let router = create_router(state);
    let router = Router::new()
        .route("/api/v1/tsigkeys", get(api::list_keys))
//        .route("/api/v1/tsigkeys/create", post(api::create_key))
//        .route("/api/v1/tsigkeys/:keyid",
//               put(api::create_key).delete(api::delete_key).get(api::list_key))
        .with_state(Arc::new(state));
    axum::Server::bind(&addr)
        .serve(router.into_make_service())
        .await.unwrap();
}

//fn create_router(state: AppState) -> Result<Router, Box<Error>> {
//    Ok(router)
//}
